"""Implementation of the 3D DistMesh algorithm."""

from __future__ import annotations

from typing import Any

import numpy as np
from scipy.spatial import Delaunay  # type: ignore

from distmeshpy._typing import Array, Float, Function, Triangulation
from distmeshpy.core import apply_force_3d, tetrahedron_find_edges
from distmeshpy.utils.config import DistMeshConfig, default_3d_config


# TODO: add loggers
def distmesh3d(
    dist_func: Function,
    size_func: Function,
    edge_size: Float,
    bounding_box: tuple[tuple[Float, Float], tuple[Float, Float]],
    fixed_points: Array | None = None,
    config: DistMeshConfig | None = None,
    **kwargs: dict[str, Any],
) -> tuple[Array, Triangulation]:
    """
    Generate a 3D tetrahedral mesh using the DistMesh algorithm.

    Specialized function of ``distmeshnd.m`` in the original MATLAB code. This function
    only works for 3D tetrahedral meshes, while the original function is generalized for
    simplexes in N dimensions.

    :param dist_func: Distance function.
    :type dist_func: Function
    :param size_func: Scaled edge length function.
    :type size_func: Function
    :param edge_size: Initial edge size.
    :type edge_size: Float
    :param fixed_points: Array of points to be fixed in the mesh.
    :type fixed_points: Array, None
    :param bounding_box: Bounding box for the mesh in the form
        ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
    :type bounding_box: tuple[tuple[Float, Float], tuple[Float, Float], tuple[Float, Float]]
    :param config: Configuration object for DistMesh. Defaulted to default_3d_config().
    :type config: DistMeshConfig, None
    :param kwargs: Additional arguments to be passed to the distance and size functions.
    :type kwargs: dict[str, Any]
    :return: Mesh points and triangulation
    """
    # 0. Define config and tolerance parameters
    if config is None:
        config = default_3d_config()
    eps = config.machine_eps * edge_size
    geo_eps = config.geometry_eps * edge_size
    step_tol = config.step_tol * edge_size
    mesh_tol = config.mesh_tol * edge_size

    # 1. Initialize the points
    grid_x, grid_y, grid_z = np.meshgrid(
        np.arange(bounding_box[0][0], bounding_box[0][1] + eps, edge_size),
        np.arange(bounding_box[1][0], bounding_box[1][1] + eps, edge_size),
        np.arange(bounding_box[2][0], bounding_box[2][1] + eps, edge_size),
    )
    grid_x = np.transpose(grid_x, (1, 0, 2))
    grid_y = np.transpose(grid_y, (1, 0, 2))
    grid_z = np.transpose(grid_z, (1, 0, 2))
    points = np.vstack((grid_x.ravel(), grid_y.ravel(), grid_z.ravel())).T

    # 2. Remove points outside the boundary via the rejection method
    points = points[np.where(dist_func(points, **kwargs) < geo_eps)[0], :]
    r0 = size_func(points, **kwargs) ** 3
    rng = np.random.default_rng(seed=kwargs.get("seed"))
    points = points[
        np.where(rng.uniform(size=points.shape[0]) < np.amin(r0) / r0)[0], :
    ]
    n_fix = 0
    if fixed_points is not None:
        fixed_pts = np.unique(fixed_points, axis=0)
        n_fix = fixed_pts.shape[0]
        to_remove = []
        for i in range(fixed_points.shape[0]):
            to_remove += np.where(
                np.all(np.isclose(points, fixed_points[i, :]), axis=1)
            )[0].tolist()
        points = np.delete(points, to_remove, axis=0)
        points = np.vstack((fixed_pts, points))

    # Main loop (time steps)
    count = 0
    d_count = 0
    step_change = config.step_tol + 1.0
    previous_points = np.inf

    bars: Triangulation = np.zeros((2,), dtype=np.int32)
    triangulation: Triangulation = np.zeros((3,), dtype=np.int32)

    while step_change > step_tol and count < config.max_iter:
        count += 1

        # 3. Re-triangulation by the Delaunay algorithm
        if np.amax(np.linalg.norm(points - previous_points, axis=1)) >= mesh_tol:
            d_count += 1

            previous_points = points.copy()  # type: ignore
            tri = Delaunay(points)
            triangulation = tri.simplices.astype(np.int32, copy=False)
            mid_points = np.mean(points[triangulation], axis=1)
            triangulation = triangulation[
                np.where(dist_func(mid_points, **kwargs) < -geo_eps)[0], :
            ]

            # 4. Describe each bar by a unique pair of nodes
            bars = tetrahedron_find_edges(triangulation)

        # 5. Move mesh points based on bar lengths L and forces F
        bar_vec = points[bars[:, 0], :] - points[bars[:, 1], :]
        bar_lengths = np.linalg.norm(bar_vec, axis=1).reshape((-1, 1))
        bar_size_vals = size_func(
            0.5 * (points[bars[:, 0], :] + points[bars[:, 1], :]), **kwargs
        )
        desired_lengths = (
            config.internal_pressure
            * bar_size_vals
            * (np.sum(bar_lengths**3) / np.sum(bar_size_vals**3)) ** (1 / 3)
        ).reshape((-1, 1))

        # Density control. Remove points that are too close to each other
        if count % config.density_ctrl_freq == 0:
            close_bars = desired_lengths > 2 * bar_lengths
            if np.any(close_bars):
                points_to_remove = np.setdiff1d(
                    bars[np.where(close_bars)[0], :].ravel(), np.arange(0, n_fix)
                )
                points = np.delete(points, np.unique(points_to_remove), axis=0)
                previous_points = np.inf  # type: ignore
                continue

        force_vectors = (
            np.maximum(desired_lengths - bar_lengths, 0.0) / bar_lengths
        ) * bar_vec
        force_on_bars = np.zeros_like(points)
        apply_force_3d(force_on_bars, force_vectors, bars)

        force_on_bars[:n_fix, :] = 0.0
        points += config.euler_step * force_on_bars

        # 6. Bring outside points back to the boundary
        dist = dist_func(points, **kwargs)
        out_points = dist > 0
        out_points[:n_fix] = False
        if np.any(out_points):
            tmp = (
                np.vstack(
                    (
                        points[out_points, 0] + eps,
                        points[out_points, 1],
                        points[out_points, 2],
                    )
                ).T
            ).astype(np.float64)
            dist_grad_x = (dist_func(tmp, **kwargs) - dist[out_points]) / eps
            tmp = (
                np.vstack(
                    (
                        points[out_points, 0],
                        points[out_points, 1] + eps,
                        points[out_points, 2],
                    )
                ).T
            ).astype(np.float64)
            dist_grad_y = (dist_func(tmp, **kwargs) - dist[out_points]) / eps
            tmp = (
                np.vstack(
                    (
                        points[out_points, 0],
                        points[out_points, 1],
                        points[out_points, 2] + eps,
                    )
                ).T
            ).astype(np.float64)
            dist_grad_z = (dist_func(tmp, **kwargs) - dist[out_points]) / eps
            dist_grad = np.vstack((dist_grad_x, dist_grad_y, dist_grad_z)) / (
                dist_grad_x**2 + dist_grad_y**2 + dist_grad_z**2
            )
            points[out_points, :] -= (dist[out_points] * dist_grad).T

        step_change = np.sqrt(config.euler_step) * np.amax(
            np.linalg.norm(force_on_bars[dist < -geo_eps, :], axis=1)
        )

    return points, triangulation



"""
Provides access to 2D, 3D, and surface mesh generation functions.

These functions mirror the functionality of the original DistMesh MATLAB
program at http://persson.berkeley.edu/distmesh/, with the exception of
`distmesh3d` which in the original code is generalized as `dishmeshnd`.
"""

from distmeshpy.mesh.distmesh2d import distmesh2d
from distmeshpy.mesh.distmesh3d import distmesh3d
from distmeshpy.mesh.distmeshsurface import distmeshsurface

__all__ = ["distmesh2d", "distmesh3d", "distmeshsurface"]



"""Wrapper utility functions for tetrahedron meshes, implemented in C++."""

from distmeshpy._typing import Array, Triangulation

from ._tetrahedron_utils import (
    _tetrahedron_find_edges,
    _tetrahedron_find_surface,
)


def tetrahedron_find_edges(f2v: Triangulation) -> Triangulation:
    """
    Find edges of a tetrahedron mesh.

    :param f2v: (Ne, 4) The tetrahedron mesh.
    :type f2v: Triangulation
    :return: The edges of the tetrahedron mesh.
    :rtype: Triangulation
    """
    return _tetrahedron_find_edges(f2v)


def tetrahedron_find_surface(pts: Array, f2v: Triangulation) -> Triangulation:
    """
    Find the surface of a tetrahedron mesh.

    :param pts: (N, 3) The vertex coordinates of the tetrahedron mesh.
    :type pts: Array
    :param f2v: (Ne, 4) The tetrahedron mesh.
    :type f2v: Triangulation
    :return: The surface of the tetrahedron mesh.
    :rtype: Triangulation
    """
    return _tetrahedron_find_surface(pts, f2v)



"""Implementation of distmesh algorithm for surfaces in 3D space."""

from __future__ import annotations

from typing import Any

import numpy as np
from mcubes import marching_cubes  # type: ignore

from distmeshpy._typing import Array, Float, Function, Triangulation
from distmeshpy.core import TriangleConnectivity, apply_force_3d, triangle_find_edges
from distmeshpy.utils.config import DistMeshConfig, default_surface_config


def distmeshsurface(
    dist_func: Function,
    size_func: Function,
    edge_size: Float,
    bounding_box: tuple[tuple[Float, Float], tuple[Float, Float], tuple[Float, Float]],
    config: DistMeshConfig | None = None,
    **kwargs: dict[str, Any],
) -> tuple[Array, Triangulation]:
    """
    Generate 3D surface mesh.

    Mirror function of ``distmeshsurface.m`` in the original MATLAB code. From original
    author: "Doesn't work too well for nonuniform size functions. Need a better initial
    grid or density control. Also no support for fixed points, again because of the
    initial grid."

    :param dist_func: Distance function.
    :type dist_func: Function
    :param size_func: Scaled edge length function.
    :type size_func: Function
    :param edge_size: Initial edge size.
    :type edge_size: Float
    :param bounding_box: Bounding box for the mesh in the form
        ((x_min, x_max), (y_min, y_max), (z_min, z_max)).
    :type bounding_box: tuple[tuple[Float, Float], tuple[Float, Float], tuple[Float, Float]]
    :param config: Configuration object for DistMesh. Defaulted to default_surface_config().
    :type config: DistMeshConfig, None
    :param kwargs: Additional arguments to be passed to the distance and size functions.
    :type kwargs: dict[str, Any]
    :return: Mesh points and triangulation
    """
    # 0. Define config and tolerance parameters
    if config is None:
        config = default_surface_config()
    eps = config.machine_eps * edge_size
    step_tol = config.step_tol * edge_size
    mesh_tol = config.mesh_tol * edge_size

    # 1. Initialize the points
    grid_x, grid_y, grid_z = np.meshgrid(
        np.arange(bounding_box[0][0], bounding_box[0][1] + eps, edge_size),
        np.arange(bounding_box[1][0], bounding_box[1][1] + eps, edge_size),
        np.arange(bounding_box[2][0], bounding_box[2][1] + eps, edge_size),
    )
    ny, nx, nz = grid_x.shape
    grid_x = np.transpose(grid_x, (1, 0, 2))
    grid_y = np.transpose(grid_y, (1, 0, 2))
    grid_z = np.transpose(grid_z, (1, 0, 2))
    points = np.vstack((grid_x.ravel(), grid_y.ravel(), grid_z.ravel())).T

    # iso-surface is found by the marching-cubes algorithm.
    # The code below is equivalent to the MATLAB isosurface function
    points, triangulation = marching_cubes(
        dist_func(points, **kwargs).reshape((nx, ny, nz)), 0.0
    )
    triangulation = triangulation.astype(np.int32, copy=False)
    points = points.astype(np.float64, copy=False)

    points *= edge_size
    points += np.array(
        [bounding_box[0][0], bounding_box[1][0], bounding_box[2][0]], dtype=np.float64
    )
    tri_conn = TriangleConnectivity(triangulation)

    # Main loop (time steps)
    count = 0
    d_count = 0
    step_change = config.step_tol + 1.0
    mesh_points = np.inf

    bars: Triangulation = np.zeros((2,), dtype=np.int32)

    while step_change > step_tol and count < config.max_iter:
        if np.isnan(points).any():
            message = "NaN detected in points"
            raise ValueError(message)

        count += 1
        step_points = points.copy()  # type: ignore

        # 3. Re-triangulation by the Delaunay algorithm
        if np.amax(np.linalg.norm(points - mesh_points, axis=1)) >= mesh_tol:
            d_count += 1
            mesh_points = points.copy()  # type: ignore
            tri_conn.update_connectivity(points.view(), triangulation.view())

            # 4. Describe each bar by a unique pair of nodes
            bars = triangle_find_edges(triangulation)

        # 5. Move mesh points based on bar lengths L and forces F
        bar_vec = points[bars[:, 0], :] - points[bars[:, 1], :]
        bar_lengths = np.linalg.norm(bar_vec, axis=1).reshape((-1, 1))
        bar_size_vals = size_func(
            0.5 * (points[bars[:, 0], :] + points[bars[:, 1], :]), **kwargs
        )
        desired_lengths = (
            config.internal_pressure
            * bar_size_vals
            * np.sqrt(np.sum(bar_lengths**2) / np.sum(bar_size_vals**2))
        ).reshape((-1, 1))
        force_vectors = (
            np.maximum(desired_lengths - bar_lengths, 0.0) / bar_lengths
        ) * bar_vec
        force_on_bars = np.zeros_like(points)
        apply_force_3d(force_on_bars, force_vectors, bars)
        points += config.euler_step * force_on_bars

        # 6. Bring outside points back to the boundary
        dist = dist_func(points, **kwargs)
        tmp = (
            np.vstack(
                (
                    points[:, 0] + eps,
                    points[:, 1],
                    points[:, 2],
                )
            ).T
        ).astype(np.float64)
        dist_grad_x = (dist_func(tmp, **kwargs) - dist) / eps
        tmp = (
            np.vstack(
                (
                    points[:, 0],
                    points[:, 1] + eps,
                    points[:, 2],
                )
            ).T
        ).astype(np.float64)
        dist_grad_y = (dist_func(tmp, **kwargs) - dist) / eps
        tmp = (
            np.vstack(
                (
                    points[:, 0],
                    points[:, 1],
                    points[:, 2] + eps,
                )
            ).T
        ).astype(np.float64)
        dist_grad_z = (dist_func(tmp, **kwargs) - dist) / eps
        dist_grad = np.vstack((dist_grad_x, dist_grad_y, dist_grad_z)) / (
            dist_grad_x**2 + dist_grad_y**2 + dist_grad_z**2
        )
        points -= (dist * dist_grad).T

        step_change = np.amax(np.linalg.norm(points - step_points, axis=1))

    return points, triangulation



"""Internal utility wrapper functions implemented in C++."""

from distmeshpy._typing import Array, Triangulation

from ._internal import _apply_force_2d, _apply_force_3d


def apply_force_2d(
    force_on_bar: Array, force_vector: Array, bar_index: Triangulation
) -> None:
    """
    Assemble the force vector on the trusses in 2D.

    :param force_on_bar: (Ne, 2) The force vector to assemble.
    :type force_on_bar: np.typing.NDArray[np.float64]
    :param force_vector: (Nb, 2) Force vector on each individual truss.
        Each row corresponds to the row in bar_index.
    :type force_vector: np.typing.NDArray[np.float64]
    :param bar_index: (Nb, 2) The nodal index of each individual truss.
    :type p3: np.typing.NDArray[np.int32]
    """
    _apply_force_2d(force_on_bar, force_vector, bar_index)


def apply_force_3d(
    force_on_bar: Array, force_vector: Array, bar_index: Triangulation
) -> None:
    """
    Assemble the force vector on the trusses in 3D.

    :param force_on_bar: (Ne, 3) The force vector to assemble.
    :type force_on_bar: np.typing.NDArray[np.float64]
    :param force_vector: (Nb, 3) Force vector on each individual truss.
        Each row corresponds to the row in bar_index.
    :type force_vector: np.typing.NDArray[np.float64]
    :param bar_index: (Nb, 2) The nodal index of each individual truss.
    :type p3: np.typing.NDArray[np.int32]
    """
    _apply_force_3d(force_on_bar, force_vector, bar_index)



"""Predefined distance functions."""

import numpy as np

from distmeshpy._typing import Array, Float


def drectangle(p: Array, x1: Float, x2: Float, y1: Float, y2: Float) -> Array:
    """
    Evaluate the signed distance from a rectangle.

    :param p: (N, 2) The points to evaluate the signed distance at.
    :type p: Array
    :param x1: The x-coordinate of the left edge of the rectangle.
    :type x1: Float
    :param x2: The x-coordinate of the right edge of the rectangle.
    :type x2: Float
    :param y1: The y-coordinate of the bottom edge of the rectangle.
    :type y1: Float
    :param y2: The y-coordinate of the top edge of the rectangle.
    :type y2: Float
    :return: The signed distance from the rectangle.
    :rtype: Array
    """
    return -np.min(
        np.vstack((-y1 + p[:, 1], y2 - p[:, 1], -x1 + p[:, 0], x2 - p[:, 0])), axis=0
    )


def dcircle(p: Array, x0: Float, y0: Float, r: Float) -> Array:
    """
    Evaluate the signed distance from a circle.

    :param p: (N, 2) The points to evaluate the signed distance at.
    :type p: Array
    :param x0: The x-coordinate of the center of the circle.
    :type x0: Float
    :param y0: The y-coordinate of the center of the circle.
    :type y0: Float
    :param r: The radius of the circle.
    :type r: Float
    :return: The signed distance from the circle.
    :rtype: Array
    """
    return np.sqrt((p[:, 0] - x0) ** 2 + (p[:, 1] - y0) ** 2) - r


def ddiff(d1: Array, d2: Array) -> Array:
    """
    Evaluate the signed distance from the difference of two shapes.

    :param d1: The signed distance from the first shape (removed from).
    :type d1: Array
    :param d2: The signed distance from the second shape (to remove).
    :type d2: Array
    :return: The signed distance from the difference of the two shapes.
    :rtype: Array
    """
    return np.maximum(d1, -d2)


def dunion(d1: Array, d2: Array) -> Array:
    """
    Evaluate the signed distance from the union of two shapes.

    :param d1: The signed distance from the first shape.
    :type d1: Array
    :param d2: The signed distance from the second shape.
    :type d2: Array
    :return: The signed distance from the union of the two shapes.
    :rtype: Array
    """
    return np.minimum(d1, d2)


def dintersect(d1: Array, d2: Array) -> Array:
    """
    Evaluate the signed distance from the intersection of two shapes.

    :param d1: The signed distance from the first shape.
    :type d1: Array
    :param d2: The signed distance from the second shape.
    :type d2: Array
    :return: The signed distance from the intersection of the two shapes.
    :rtype: Array
    """
    return np.maximum(d1, d2)


def dsphere(p: Array, xc: Float, yc: Float, zc: Float, r: Float) -> Array:
    """
    Evaluate the signed distance from a sphere.

    :param p: (N, 3) The points to evaluate the signed distance at.
    :type p: Array
    :param xc: The x-coordinate of the center of the sphere.
    :type xc: Float
    :param yc: The y-coordinate of the center of the sphere.
    :type yc: Float
    :param zc: The z-coordinate of the center of the sphere.
    :type zc: Float
    :param r: The radius of the sphere.
    :type r: Float
    :return: The signed distance from the sphere.
    :rtype: Array
    """
    return np.sqrt((p[:, 0] - xc) ** 2 + (p[:, 1] - yc) ** 2 + (p[:, 2] - zc) ** 2) - r



from typing import Any, Protocol

from numpy import float32, float64, int32
from numpy.typing import NDArray
from typing_extensions import TypeAlias

Float: TypeAlias = "float | float32 | float64"
Int: TypeAlias = "int | int32"
Array: TypeAlias = NDArray[float64]
Triangulation: TypeAlias = NDArray[int32]


class Function(Protocol):
    def __call__(self, array: Array, **kwargs: dict[str, Any]) -> Array: ...



"""Predefined size distribution functions."""

from typing import Any

import numpy as np

from distmeshpy._typing import Array


def huniform(p: Array, **kwargs: dict[str, Any]) -> Array:  # noqa: ARG001
    """Return uniform size distribution."""
    return np.ones((p.shape[0],))



"""Wrapper utility functions for triangular meshes, implemented in C++."""

from distmeshpy._typing import Array, Float, Triangulation

from ._triangle_utils import (
    _triangle_area,
    _triangle_area_all,
    _triangle_find_edges,
    _triangle_normal,
    _triangle_normal_all,
    _triangle_quality,
    _triangle_quality_all,
    _TriangleConnectivity,
)


def triangle_area(p1: Array, p2: Array, p3: Array) -> Float:
    """
    Find the area of a triangle in 3D space.

    :param p1: (3, ) First vertex.
    :type p1: Array
    :param p2: (3, ) Second vertex.
    :type p2: Array
    :param p3: (3, ) Third vertex.
    :type p3: Array
    :return: Area of the triangle
    :rtype: Float
    """
    return _triangle_area(p1, p2, p3)


def triangle_area_all(p1: Array, p2: Array, p3: Array) -> Array:
    """
    Find the area of all triangle in 3D space defined by each row of the input arrays.

    :param p1: (N, 3) First vertices.
    :type p1: Array
    :param p2: (N, 3) Second vertices.
    :type p2: Array
    :param p3: (N, 3) Third vertices.
    :type p3: Array
    :return: (N, ) Areas of the triangle
    :rtype: Array
    """
    return _triangle_area_all(p1, p2, p3)


def triangle_normal(p1: Array, p2: Array, p3: Array) -> Array:
    """
    Find the unit normal vector of a triangle in 3D space.

    The employed convention assumes counter-clockwise vertex indexing.
    Hence, the normal vector is the normalized (p2 - p1) x (p3 - p1).

    :param p1: (3, ) First vertex.
    :type p1: Array
    :param p2: (3, ) Second vertex.
    :type p2: Array
    :param p3: (3, ) Third vertex.
    :type p3: Array
    :return: (3, ) Normal vector of the triangle.
    :rtype: Array
    """
    return _triangle_normal(p1, p2, p3)


def triangle_normal_all(p1: Array, p2: Array, p3: Array) -> Array:
    """
    Find the unit normal vector of all input triangles in 3D space.

    The employed convention assumes counter-clockwise vertex indexing.
    Hence, the normal vector is the normalized (p2 - p1) x (p3 - p1).

    :param p1: (N, 3) First vertices.
    :type p1: Array
    :param p2: (N, 3) Second vertices.
    :type p2: Array
    :param p3: (N, 3) Third vertices.
    :type p3: Array
    :return: (N, 3) Normal vectors of given triangles triangle.
    :rtype: Array
    """
    return _triangle_normal_all(p1, p2, p3)


def triangle_quality(p1: Array, p2: Array, p3: Array) -> Float:
    """
    Find the mesh quality of a triangle in 3D space.

    :param p1: (3, ) First vertex.
    :type p1: Array
    :param p2: (3, ) Second vertex.
    :type p2: Array
    :param p3: (3, ) Third vertex.
    :type p3: Array
    :return: Quality of the triangle
    :rtype: Float
    """
    return _triangle_quality(p1, p2, p3)


def triangle_quality_all(p1: Array, p2: Array, p3: Array) -> Array:
    """
    Find the mesh quality of all input triangles in 3D space.

    :param p1: (N, 3) First vertices.
    :type p1: Array
    :param p2: (N, 3) Second vertices.
    :type p2: Array
    :param p3: (N, 3) Third vertices.
    :type p3: Array
    :return: (N, ) Qualities of all input triangles.
    :rtype: Array
    """
    return _triangle_quality_all(p1, p2, p3)


def triangle_find_edges(f2v: Triangulation) -> Triangulation:
    """
    Find all edges of a triangular mesh.

    :param p1: (Ne, 3) The triangular mesh.
    :type p1: Triangulation
    :return: (N, 2) Edges of the mesh.
    :rtype: Float
    """
    return _triangle_find_edges(f2v)


class TriangleConnectivity(_TriangleConnectivity):
    """An auxiliary class to handle the connectivity of a triangular mesh."""

    def __init__(self, f2v: Triangulation) -> None:
        """
        Initialize a TriangleConnectivity object.

        A triangular mesh is required to initialize a Triangulation object.
        The mesh is processed to find all shared edges along with the connecting
        triangles. After initialization, the f2v object should not be modified
        in between calls to the update_connectivity method, as the internally
        stored connectivity information may become invalid.

        This class is used internally by the distmesh2d and
        distmeshsurface functions. Typically, users should not find the need to
        use this class directly.

        :param f2v: Input triangular mesh.
        :type f2v: Triangulation
        """
        super().__init__(f2v)

    def update_connectivity(self, points: Array, f2v: Triangulation) -> None:
        """
        Evaluate the quality of meshes and attempt to improve them via edge flipping.

        This functionality of this method mirrors that of "trisurfupd" in the original
        MATLAB implementation.

        :param points: Nodal coordinates of the mesh.
        :type points: Array
        :param f2v: The triangular mesh.
        :type f2v: Triangulation
        """
        self._update_connectivity(points, f2v)



"""
Configuration class containing mesh parameters for distmesh.

Recommended sets of default parameters generated via the provided
factory functions. These parameters suffice for most use cases.
Custom parameters, such as maximum number of iterations, should be
specified by creating a new instance of DistMeshConfig.
"""

from dataclasses import dataclass

import numpy as np

from distmeshpy._typing import Float, Int

MACHINE_EPS_SQRT = np.sqrt(np.finfo(np.float64).eps)


@dataclass
class DistMeshConfig:
    """
    Configuration class for DistMesh.

    This class holds parameters used for mesh generation, such as tolerances,
    step sizes, and machine precision.
    """

    internal_pressure: Float = 1.2
    step_tol: Float = 1e-3
    mesh_tol: Float = 0.1
    euler_step: Float = 0.2
    machine_eps: Float = MACHINE_EPS_SQRT
    density_ctrl_freq: Int = 30
    max_iter: Int = 4000
    geometry_eps: Float = 1e-3


def default_2d_config() -> DistMeshConfig:
    """
    Generate default configuration for 2D meshes.

    :return: Default configuration object for 2D meshes.
    :rtype: DistMeshConfig
    """
    return DistMeshConfig()


def default_surface_config() -> DistMeshConfig:
    """
    Generate default configuration for 3D surface meshes.

    :return: Default configuration object for 3D surface meshes.
    :rtype: DistMeshConfig
    """
    return DistMeshConfig(step_tol=np.float64(1e-4))


def default_3d_config() -> DistMeshConfig:
    """
    Generate default configuration for 3D meshes.

    :return: Default configuration object for 3D meshes.
    :rtype: DistMeshConfig
    """
    return DistMeshConfig(
        internal_pressure=np.float64(1.1),
        geometry_eps=np.float64(0.1),
        euler_step=np.float64(0.1),
    )



import numpy
import pybind11
from setuptools import Extension, setup


def create_extension(name: str, sources: list[str]) -> None:
    return Extension(
        name,
        sources=sources,
        include_dirs=["distmeshpy/core/src", pybind11.get_include(), numpy.get_include()],
        language="c++",
        extra_compile_args=["-O3", "-march=native", "-std=c++11"],
    )


ext_modules = [
    create_extension(
        "distmeshpy.core._triangle_utils", ["distmeshpy/core/src/triangle_utils.cpp"]
    ),
    create_extension(
        "distmeshpy.core._tetrahedron_utils", ["distmeshpy/core/src/tetrahedron_utils.cpp", "distmeshpy/core/src/triangle_utils.cpp"]
    ),
    create_extension("distmeshpy.core._internal", ["distmeshpy/core/src/internal.cpp"]),
]

setup(ext_modules=ext_modules)
